'''Generate vehicle trips for SUMO simulation.
Generation of headway according to negative exponential distribution.
https://nptel.ac.in/courses/105101008/downloads/cete_13.pdf
'''

import math
import random

random.seed()


def generate_times(vehicles_per_hour, simulation_time=3600, min_val=1.2):
    tsim = 0
    # mu: average headway [s]
    mu = 3600 / vehicles_per_hour - min_val
    while True:
        r = random.random()
        t = -mu * math.log(r) + min_val
        tsim += t
        if tsim > simulation_time:
            break
        yield tsim


def create_trips(prob, from_, to, flow, **kwargs):
    trips = []
    types = list(prob.keys())
    cprob = []
    csum = 0
    for v in prob.values():
        csum += v
        cprob.append(csum)
    if csum != 1:
        cprob = [x / csum for x in cprob]
    stat = {k: 0 for k in types}
    for t in generate_times(flow, min_val=1.2):
        ti = int(round(t))
        r = random.random()
        i = len([x for x in cprob if r > x])
        stat[types[i]] += 1
        trip = {'depart': ti, 'type': types[i], 'from': from_, 'to': to}
        trip.update(kwargs)
        trips.append(trip)
    # print(stat)
    return trips


def write_trips(filename, trips, **kwargs):
    s_trips = sorted(trips, key=lambda x: x['depart'])
    with open(filename, 'w') as fout:
        fout.write('<trips>\n')
        for i, trip in enumerate(s_trips):
            trip.update(kwargs)
            options = ['{}="{}"'.format(k, v) for k, v in trip.items()]
            fout.write('\t<trip id="{:05}" {} />\n'.format(i, ' '.join(options)))
        #     fout.write('\t<trip id="{:05}" depart="{}" from="{}" to="{}" type="{}" />\n'
        #             .format(i, trip['depart'], trip['from'], trip['to'], trip['type']))
        fout.write('</trips>\n')


if __name__ == '__main__':
    # for t in generate_times(900, min_val=1.2):
    #     print(int(round(t)), end=' ')
    prob = {'psg': 0.8, 'trk': 0.1, 'del': 0.07, 'bus': 0.03}
    #tripsEW = create_trips(prob, 'Estart', 'Wout', 600, color='0,162,232')
    
    flow_scaler = 4
    
    total_trips = []
    total_trips += create_trips(prob, 'Nstart', 'Eout', flow_scaler*100, color='128,128,12')
    # total_trips += create_trips(prob, 'Estart', 'Nout', flow_scaler*100, color='255,255,255')
    
    # tripsNE = create_trips(prob, 'Nstart', 'Eout', flow_scaler*100, color='128,128,12')
    # tripsEN = create_trips(prob, 'Estart', 'Nout', flow_scaler*100, color='255,255,255')
    # tripsEW = create_trips(prob, 'Estart', 'Wout', flow_scaler*400, color='239,7,7')
    # tripsWE = create_trips(prob, 'Wstart', 'Eout', flow_scaler*400, color='255,119,199')
    # tripsN2W = create_trips(prob, 'N2start', 'Wout', flow_scaler*400, color='255,1,199')
    # tripsN3W = create_trips(prob, 'Wstart', 'N2out', flow_scaler*400, color='32,128,32')
    #tripsN2W = create_trips(prob, 'N2start', 'Wout', 400, color='64,256,0')
    #tripsN3W = create_trips(prob, 'N3start', 'Wout', 400, color='0,128,0')	
    #tripsNE2 = create_trips(prob, 'N2start','Eout', 400, color='256,32,0')
    #tripsNE3 = create_trips(prob, 'N3start', 'Eout', 400, color='16,32,64')
	
	
	#tripsNE = create_trips(prob, 'Nstart', 'Eout', 600, color='0,0,128')
    #tripsEN = create_trips(prob, 'Estart', 'Nout', 600, color='0,128,0')
    #tripsSE = create_trips(prob, 'Sstart', 'Eout', 600, color='0,255,255')
    #tripsES = create_trips(prob, 'Estart', 'Sout', 600, color='255,0,255')
    
    #tripsNW = create_trips(prob, 'Nstart', 'Wout', 300, color='255,255,0')
	
#tripsEW + tripsWE + tripsNS + tripsSN + tripsNE + tripsEN + tripsSE + tripsES + tripsWN +tripsNW
    write_trips('trips.xml', total_trips )# tripsN2W + tripsN3W + tripsNE2 +tripsNE3 )
    
