from __future__ import absolute_import
from __future__ import print_function

import numpy as np
import os
import sys
import pandas
import matplotlib.pyplot as plt
import pickle
import time

# we need to import python modules from the $SUMO_HOME/tools directory
if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)
else:
    sys.exit("please declare environment variable 'SUMO_HOME'")

from sumolib import checkBinary  # noqa
import traci  # noqa

# Decision Model --------------------------------------------------

def minimal_distance_to_velocity(minimal_distance, pavement_factor):
    a = 1/(2*pavement_factor)
    
    # b = 0.85 # reaction time in seconds
    # b = 1.15 # reaction time in seconds
    b = 1.50 # reaction time in seconds
    
    c = -minimal_distance
    
    solutions = np.roots([a,b,c])
    
    if solutions[0] >= 0:
        desired_velocity_ms = solutions[0]
    else:
        desired_velocity_ms = solutions[1]
    
    return desired_velocity_ms

def suggest_velocity_v3 (mean_distance_between_vehicles, vehicle_velocity_ms,pavement_factor=7.5,minimal_visibility_distance=2000):
    
    critical_distance       = minimal_visibility_distance # chwilowo odpuszczamy heurystykę, bo daje ona dziwne, zaniżone prędkości
    
    recommended_velocity    = minimal_distance_to_velocity(mean_distance_between_vehicles, pavement_factor) # m/s
    
    critical_velocity       = minimal_distance_to_velocity(critical_distance, pavement_factor) # m/s
    
    recommended_velocity = np.min([critical_velocity, recommended_velocity]) # m/s
    return recommended_velocity # m/s
    
# Tracking Code ------------------------------------------------------

class ParameterTracker:
    def __init__(self, tracker_name, parameter_name):
        self.tracker_name   = tracker_name
        self.parameter_name = parameter_name
    
class DetectorMonitor:
    def __init__(self, detector_id, traffic_ctrl_model_on):
        self.detector_id                    = detector_id
        self.det_arr                        = []
        self.intensity_estim_buffer         = [0]*60
        self.intensity_estimate             = []
        self.velocities_buffer              = []
        self.car_pass_detected_v            = False
        self.spdlm_arr                      = []
        self.detector_lenght_estim_buffer   = []
        self.traffic_ctrl_model_on  = traffic_ctrl_model_on      
        self.mean_distance          = np.inf
        self.mean_distance_track    = []
        self.traffic_density_track  = []
        self.mean_velocity_track    = []
        
        self.SPEED_LIMIT = 90*10/36 # [m/s]
        self.sugested_velocity_6 = self.SPEED_LIMIT
        
    def tracked_stats(self):
        output_structure = {}
        output_structure.update({'mean_distance':self.mean_distance_track})
        output_structure.update({'traffic_density':self.traffic_density_track})
        return output_structure
        
    def car_pass_detected(self):
        return self.car_pass_detected_v
    
    def tick(self):
        detection_time = traci.inductionloop.getTimeSinceDetection(self.detector_id)
        self.det_arr.append(detection_time)
        
        
        self.car_pass_detected_v = False
        if len(self.det_arr) >= 2: # operating on differentials needs at leas 2 samples in det_arr
            self.intensity_estim_buffer.pop(0) # the buffer is used to track instantaneous traffic intensity
            if (self.det_arr[-1]-self.det_arr[-2]) < 0: # detection on negative edge 
                self.car_pass_detected_v = True
                self.intensity_estim_buffer.append(1) # if car pass - append 1 to intensity tracking buffer
                vehicle_id = traci.inductionloop.getLastStepVehicleIDs(self.detector_id)
                vehicle_velocity = traci.vehicle.getSpeed(vehicle_id[0]) # velocity on detector-cross event
                self.velocities_buffer.append(vehicle_velocity)
            else:
                self.intensity_estim_buffer.append(0) # if no pass - append 0
            self.intensity_estimate.append(np.sum(self.intensity_estim_buffer)) # by summing over 60 samples(1 second long ticks) long buffer you obtain intensity in vehicles/minute
            
        if self.car_pass_detected():
            print('-------------------------------------------')
            print(self.detector_id, 'id: ',vehicle_id[0])
            # log the intensity only on value change (to avoid cluttering of the log)
            if len(self.intensity_estimate)>2:
                if self.intensity_estimate[-1] != self.intensity_estimate[-2]:
                    print(self.detector_id, 'traffic intensity: %3.3f [veh./min]'%self.intensity_estimate[-1])
                    # on intensity change we also will log new mean distance 
                    # we assume vehicles are 5 meters long
                    traffic_density = (self.intensity_estimate[-1]/60)/np.mean(self.velocities_buffer)
                    
                    if len(self.detector_lenght_estim_buffer) == 0:
                        mean_vehicle_lenght = 5
                    else:
                        mean_vehicle_lenght = np.mean(self.detector_lenght_estim_buffer)
                       
                    self.mean_distance = 1/traffic_density - mean_vehicle_lenght
                    self.mean_distance_track.append(self.mean_distance)
                    self.traffic_density_track.append(traffic_density)
                    print (self.detector_id, 'mean vehcile lenght: %3.3f [m]' %mean_vehicle_lenght )
            
            vehicle_id = traci.inductionloop.getLastStepVehicleIDs(self.detector_id)
            vehicle_lenght = traci.vehicle.getLength(vehicle_id[0])
            vehicle_velocity = traci.vehicle.getSpeed(vehicle_id[0])
            self.detector_lenght_estim_buffer.append(vehicle_lenght)
            print (self.detector_id, 'vehcile velocity: %3.3f [m/s]' %vehicle_velocity )
            
            if len(vehicle_id) == 1:
                if len(self.velocities_buffer):
                    mean_velocity = np.mean(self.velocities_buffer)
                else:
                    mean_velocity = self.SPEED_LIMIT
                
                self.mean_velocity_track.append(mean_velocity)
                
                # TOGGLE FOR MODEL FUNCTIONING
                if self.traffic_ctrl_model_on:
                    # pavement_factor = 7.5
                    pavement_factor = 1.5
                    sugested_velocity_6 = suggest_velocity_v3(self.mean_distance, mean_velocity, pavement_factor,1500) # dist betw. vehs, vehs vel.,pavement factor, visib distance
                    sugested_velocity_6 = np.min([self.SPEED_LIMIT, sugested_velocity_6])
                    print(self.detector_id, 'md: %3.3f mv: %3.3f--->suggesting velocity: %3.3f [m/s] %3.3f [km/h] '%(self.mean_distance,mean_velocity,sugested_velocity_6,sugested_velocity_6*3.6))
                    traci.vehicle.setMaxSpeed(vehicle_id[0],sugested_velocity_6)
                    self.spdlm_arr.append(sugested_velocity_6)
        else:
            if len(self.mean_distance_track) == 0: self.mean_distance_track = [0]
            if len(self.traffic_density_track) == 0: self.traffic_density_track = [np.inf]
            self.mean_distance_track.append(self.mean_distance_track[-1])
            self.traffic_density_track.append(self.traffic_density_track[-1])

class LaneMonitor:
    def __init__(self, lane_name):
        self.lane_name      = lane_name
        self.data_dict      = {}
        self.data_dict.update({'C02_E4':[]})
        self.data_dict.update({'C0_E4':[]})
        self.data_dict.update({'Fuel_E4':[]})
        self.data_dict.update({'Electr_E4':[]})
        self.data_dict.update({'HCE_E4':[]})
        self.data_dict.update({'Noise_E4':[]})
        self.data_dict.update({'Pmx_E4':[]})
    
    def tick(self):
        self.data_dict['C02_E4'].append(traci.lane.getCO2Emission(self.lane_name))
        self.data_dict['C0_E4'].append(traci.lane.getElectricityConsumption(self.lane_name))
        self.data_dict['Fuel_E4'].append(traci.lane.getFuelConsumption(self.lane_name))
        self.data_dict['Electr_E4'].append(traci.lane.getHCEmission(self.lane_name))
        self.data_dict['HCE_E4'].append(traci.lane.getNOxEmission(self.lane_name))
        self.data_dict['Noise_E4'].append(traci.lane.getNoiseEmission(self.lane_name))
        self.data_dict['Pmx_E4'].append(traci.lane.getPMxEmission(self.lane_name))
        
    def tracked_stats(self):
        return self.data_dict
        