import time
import numpy as np
import os
import copy as cp
import matplotlib.pyplot as plt
import pandas as pd
from pandas.plotting import parallel_coordinates, andrews_curves
import matplotlib.pyplot as plt
import scipy.stats as st
import scikit_posthocs as sp

# an adapter for the multivariate test procedure
# types of multivariate tests to use:
#   st.kruskal
#   st.f_oneway

def df_multivariate_test(answers, mtest):
    test_func = None
    if mtest == 'kruskal': test_func = st.kruskal
    if mtest == 'ANOVA': test_func = st.f_oneway
    if test_func is None: raise RuntimeError('Bad mtest specified.')
    
    observation_vectors = []
    for col_name in answers.columns:
        observation_vectors.append(answers[col_name].values)
    if mtest=='kruskal':
        print(test_func(*observation_vectors, nan_policy = 'omit'))
    if mtest=='ANOVA':
        print(test_func(*observation_vectors))

# an adapter for post-hoc procedure
def make_postohcs(data, posthoc_test):
    posthoc_func = None
    if posthoc_test == 'Dunn': posthoc_func = sp.posthoc_dunn
    if posthoc_test == 'TukeyHSD': posthoc_func = sp.posthoc_tukey_hsd
    if posthoc_func is None: raise RuntimeError('Bad mtest specified.')
    
    posthoc_table = posthoc_func(data.T.values)
    
    cmp_with_ref = posthoc_table.to_numpy().tolist()[-1]
    p_values = []
    for name, ref_result in zip(data.columns.values, cmp_with_ref):
        string_to_disp = '%50s %.2f'%(name, ref_result)
        if ref_result<0.05:
            string_to_disp += ' *'
        print(string_to_disp)
        p_values.append(ref_result)
    return np.array(p_values)

# -------------------------------------- #
# decorator function for a boxplot drawing method
def boxplot(df_table, p_values = None, significance=0.05, tilt_x=False, test_invert = False, xlabel=None, ylabel=None, title=None):

    df_table = df_table.drop(columns=['ref'])
    
    # make a new plot
    plt.figure()
    bp = plt.boxplot(df_table.transpose(), patch_artist=True)
    plt.axhline(y=0.0, color='r', linestyle='-', linewidth=2)
    
    # Gray out values which are not significant.
    gray_out = None
    if p_values is not None:
        if not test_invert:
            gray_out = np.argwhere(p_values<significance)
        else:
            gray_out = np.argwhere(p_values>significance)
    
    # plot median indicating bars in red with thicker line.
    for median_bar in bp['medians']:
        median_bar.set_linewidth(2)
        median_bar.set_color('red')
    
    # Gray out values which should be grayed out according to the code above
    if gray_out is not None:
        for key, elements in bp.items():
            if key == 'medians': continue # do not affect medians
            for i, element in enumerate(elements):
            
                # we need to calculate index in special way, because we have to repeat each index values
                # due to repetitions in elements vector.
                idx = int(np.floor(i/(len(elements)/float(len(df_table.columns)))))
                
                if idx in gray_out:
                    element.set(color='Gray')
                else:
                    element.set(color='Orange')
    
    # Display information regarding p-values, where applicable.
    if p_values is not None:
        for i,box in enumerate(bp['boxes']):
            if i >= len(p_values):   
                break
            box_vertices = box.get_path().vertices
            
            text_x = (np.max(box_vertices[:-1,0]) + np.min(box_vertices[:-1,0]))/2
            text_y = (np.max(box_vertices[:-1,1]) + np.min(box_vertices[:-1,1]))/2
            
            p_value = p_values[i]
            
            # If p-value is too small - hide it and mark it as smaller than 10^-3
            if p_value < 1e-3:
                display_text = r'$\mathbf{<10^{-3}}$'
            else:
                display_text = r'$\mathbf{%.3f}$'%(p_value)
                
            plt.text(text_x,text_y,display_text, weight='bold',horizontalalignment='center', color='black', fontsize=10)
    
    if xlabel is not None: plt.xlabel(xlabel, fontsize=18)
    if ylabel is not None: plt.ylabel(ylabel, fontsize=18)
    if title is not None: plt.title(title)
    
    # set names of parameters on the plot as ones from the data frame
    plt.gca().set_xticklabels(df_table.columns)
    
    # tilt the labels to fit them to the plot
    if tilt_x:
        plt.setp(plt.gca().xaxis.get_majorticklabels(), rotation=-30, ha="left" )
    plt.gcf().subplots_adjust(bottom=0.15)
    
    plt.gca().tick_params(axis='both', which='major', labelsize=16)
    plt.gca().tick_params(axis='both', which='minor', labelsize=16)
    plt.grid()
    
    # plt.gca().set_ylim([-2.6,0.3])